/* Copyright (C) |Meso|Star> 2015-2018 (contact@meso-star.com)
 *
 * This software is a library whose purpose is to generate [pseudo] random
 * numbers and random variates.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "ssp_rng_c.h"

#include <rsys/dynamic_array_char.h>
#include <rsys/mem_allocator.h>

/* disable some warnings in Random123 includes */
#ifdef COMPILER_GCC
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wconversion" /* unsafe conversion */
  #pragma GCC diagnostic ignored "-Wshadow"
  #pragma GCC diagnostic ignored "-Wunused-parameter" /* in Random123/aes.h */
#elif defined(COMPILER_CL)
  #pragma warning(push)
  #pragma warning(disable:4100) /* unreferenced formal parameter */
  #pragma warning(disable:4127) /* conditional expression is constant */
  #pragma warning(disable:4512) /* assignment operator could not be generated */
  #pragma warning(disable:4521) /* multiple copy constructors specified */
#endif

#include <Random123/conventional/Engine.hpp>
#include <Random123/threefry.h>
#ifdef WITH_R123_AES
  #include <Random123/aes.h>
#endif

#ifdef COMPILER_GCC
  #pragma GCC diagnostic pop
#elif defined(COMPILER_CL)
  #pragma warning(pop)
#endif

#include <sstream>
#include <cstring>
#include <limits>

/*******************************************************************************
 * KISS PRNG
 ******************************************************************************/
struct rng_kiss { uint32_t x, y, z, c; };

static res_T
rng_kiss_set(void* data, const uint64_t seed)
{
  struct rng_kiss* kiss = (struct rng_kiss*)data;
  RAN_NAMESPACE::mt19937 rng_mt(uint32_t(seed % UINT32_MAX));
  ASSERT(kiss);
  kiss->x = (uint32_t)rng_mt();
  do {
    kiss->y = (uint32_t)rng_mt();
  } while(kiss->y == 0); /* Y must be != 0 */
  kiss->z = (uint32_t)rng_mt();
  /* Offset c by 1 to avoid z=c=0; should be less than 698769069 */
  kiss->c = (uint32_t)rng_mt() % 698769068 + 1;
  return RES_OK;
}

static uint64_t
rng_kiss_get(void* data)
{
  struct rng_kiss* kiss = (struct rng_kiss*)data;
  uint64_t t;
  ASSERT(kiss);

  kiss->x = 314527869 * kiss->x + 1234567;
  kiss->y ^= kiss->y << 5;
  kiss->y ^= kiss->y >> 7;
  kiss->y ^= kiss->y << 22;
  t = 4294584393ULL * kiss->z + kiss->c;
  kiss->c = (uint32_t)(t >> 32);
  kiss->z = (uint32_t)(t & 0xFFFFFFFF);
  return (uint32_t)(kiss->x + kiss->y + kiss->z);
}

static res_T
rng_kiss_read(void* data, FILE* file)
{
  struct rng_kiss* rng = (struct rng_kiss*)data;
  struct darray_char buf;
  res_T res = RES_OK;
  int err;
  ASSERT(data && file);

  darray_char_init(&mem_default_allocator, &buf);
  res = darray_char_resize(&buf, 64);
  if(res != RES_OK) goto error;

  while(fgets(darray_char_data_get(&buf), (int)darray_char_size_get(&buf), file)) {
    /* Ensure that the whole line is read */
    if(strrchr(darray_char_cdata_get(&buf), '\n')) break;
    res = darray_char_resize(&buf, darray_char_size_get(&buf) * 2);
    if(res != RES_OK) goto error;
  }

  err = sscanf(darray_char_cdata_get(&buf), "%u %u %u %u",
    &rng->x, &rng->y, &rng->z, &rng->c);
  res = err == EOF || !err ? RES_IO_ERR : RES_OK;

exit:
  darray_char_release(&buf);
  return res;
error:
  goto exit;
}

static res_T
rng_kiss_write(const void* data, FILE* file)
{
  const struct rng_kiss* rng = (const struct rng_kiss*)data;
  int res;
  ASSERT(data && file);
  res = fprintf(file, "%u %u %u %u\n", rng->x, rng->y, rng->z, rng->c);
  return res < 0 ? RES_IO_ERR : RES_OK;
}

static res_T
rng_kiss_init(struct mem_allocator* allocator, void* data)
{
  (void)allocator;
  rng_kiss_set(data, 0);
  return RES_OK;
}

static void
rng_kiss_release(void* data)
{
  (void)data;
}

static double
rng_kiss_entropy(const void* data)
{
  (void)data;
  return 0.;
}

static res_T
rng_kiss_discard(void* data, uint64_t n)
{
  while (n-- > 0) {
    rng_kiss_get(data);
  }
  return RES_OK;
}

/* Exported type */
const struct ssp_rng_type ssp_rng_kiss = {
  rng_kiss_init,
  rng_kiss_release,
  rng_kiss_set,
  rng_kiss_get,
  rng_kiss_discard,
  rng_kiss_read,
  rng_kiss_write,
  rng_kiss_entropy,
  0,
  UINT32_MAX,
  sizeof(struct rng_kiss),
  16
};

/*******************************************************************************
 * C++11 PRNG
 ******************************************************************************/
template<typename RNG>
static res_T
rng_cxx_set(void* data, const uint64_t seed)
{
  RNG* rng = (RNG*)data;
  ASSERT(rng);
  rng->seed((typename RNG::result_type)seed);
  return RES_OK;
}

template<>
res_T
rng_cxx_set<RAN_NAMESPACE::random_device>(void* data, const uint64_t seed)
{
  (void) data; (void) seed;
  return RES_BAD_OP;
}

template<typename RNG>
static uint64_t
rng_cxx_get(void* data)
{
  RNG* rng = (RNG*)data;
  ASSERT(rng);
  return (*rng)();
}

template<typename RNG>
static res_T
rng_cxx_write(const void* data, FILE* file)
{
  size_t i;
  std::stringstream stream;
  RNG* rng = (RNG*)data;
  ASSERT(rng);
  stream << *rng << std::endl;
  i = fwrite(stream.str().c_str(), 1, stream.str().size(), file);
  return i == stream.str().size() ? RES_OK : RES_IO_ERR;
}

template<>
res_T
rng_cxx_write<RAN_NAMESPACE::random_device>(const void* data, FILE* file)
{
  (void) data; (void) file;
  return RES_BAD_OP;
}

template<typename RNG>
static res_T
rng_cxx_read(void* data, FILE* file)
{
  std::stringstream stream;
  char buf[512];
  char* s;
  RNG* rng = (RNG*)data;
  ASSERT(rng);
  while((s = fgets(buf, (int)sizeof(buf), file))) {
    stream << std::string(s);
    if(s[strlen(s)-1] == '\n') break;
  }
  stream >> *rng;
  return stream.fail() ? RES_IO_ERR : RES_OK;
}

template<>
res_T
rng_cxx_read<RAN_NAMESPACE::random_device>(void* data, FILE* file)
{
  (void) data; (void) file;
  return RES_BAD_OP;
}

template<typename RNG>
static res_T
rng_cxx_init(struct mem_allocator* allocator, void* data)
{
  (void)allocator;
  ASSERT(data);
  new (data) RNG;
  return RES_OK;
}

template<typename RNG>
static void
rng_cxx_release(void* data)
{
  RNG* rng = (RNG*)data;
  (void)rng;
  ASSERT(rng);
  rng->~RNG();
}

template<typename RNG>
static double
rng_cxx_entropy(const void* data)
{
  (void)data;
  return 0;
}

template<typename RNG>
res_T
rng_cxx_discard(void* data, uint64_t n)
{
  RNG* rng = (RNG*) data;
  ASSERT(rng);
  rng->discard(n);
  return RES_OK;
}

template<>
res_T
rng_cxx_discard<RAN_NAMESPACE::random_device>(void* data, uint64_t n)
{
  (void) data; (void) n;
  return RES_BAD_OP;
}

template<>
double
rng_cxx_entropy<RAN_NAMESPACE::random_device>(const void* data)
{
  const RAN_NAMESPACE::random_device* rng =
    (const RAN_NAMESPACE::random_device*)data;
  ASSERT(rng);
  return rng->entropy();
}

/* 64-bits Mersenne Twister PRNG */
const struct ssp_rng_type ssp_rng_mt19937_64 = {
  rng_cxx_init<RAN_NAMESPACE::mt19937_64>,
  rng_cxx_release<RAN_NAMESPACE::mt19937_64>,
  rng_cxx_set<RAN_NAMESPACE::mt19937_64>,
  rng_cxx_get<RAN_NAMESPACE::mt19937_64>,
  rng_cxx_discard<RAN_NAMESPACE::mt19937_64>,
  rng_cxx_read<RAN_NAMESPACE::mt19937_64>,
  rng_cxx_write<RAN_NAMESPACE::mt19937_64>,
  rng_cxx_entropy<RAN_NAMESPACE::mt19937_64>,
  RAN_NAMESPACE::mt19937_64::min(),
  RAN_NAMESPACE::mt19937_64::max(),
  sizeof(RAN_NAMESPACE::mt19937_64),
  16
};

/* 48-bits RANLUX PRNG */
const struct ssp_rng_type ssp_rng_ranlux48 = {
  rng_cxx_init<RAN_NAMESPACE::ranlux48>,
  rng_cxx_release<RAN_NAMESPACE::ranlux48>,
  rng_cxx_set<RAN_NAMESPACE::ranlux48>,
  rng_cxx_get<RAN_NAMESPACE::ranlux48>,
  rng_cxx_discard<RAN_NAMESPACE::ranlux48>,
  rng_cxx_read<RAN_NAMESPACE::ranlux48>,
  rng_cxx_write<RAN_NAMESPACE::ranlux48>,
  rng_cxx_entropy<RAN_NAMESPACE::ranlux48>,
  RAN_NAMESPACE::ranlux48::min(),
  RAN_NAMESPACE::ranlux48::max(),
  sizeof(RAN_NAMESPACE::ranlux48),
  16
};

/* random_device generator */
const struct ssp_rng_type ssp_rng_random_device = {
  rng_cxx_init<RAN_NAMESPACE::random_device>,
  rng_cxx_release<RAN_NAMESPACE::random_device>,
  rng_cxx_set<RAN_NAMESPACE::random_device>,
  rng_cxx_get<RAN_NAMESPACE::random_device>,
  rng_cxx_discard<RAN_NAMESPACE::random_device>,
  rng_cxx_read<RAN_NAMESPACE::random_device>,
  rng_cxx_write<RAN_NAMESPACE::random_device>,
  rng_cxx_entropy<RAN_NAMESPACE::random_device>,
  RAN_NAMESPACE::random_device::min(),
  RAN_NAMESPACE::random_device::max(),
  sizeof(RAN_NAMESPACE::random_device),
  16
};

/*******************************************************************************
 * Random123 Counter Based RNG
 ******************************************************************************/
typedef r123::Engine<r123::Threefry4x64> threefry_T;

#ifdef WITH_R123_AES
typedef r123::Engine<r123::AESNI4x32> aes_T;

template<>
res_T
rng_cxx_init<aes_T>(struct mem_allocator* allocator, void* data)
{
  (void) allocator;

  if (!haveAESNI()) {
    /* AES-NI instructions not available on this hardware */
    return RES_BAD_OP;
  }
  ASSERT(data);
  new (data) aes_T;
  return RES_OK;
}
#endif

/* threefry generator */
const struct ssp_rng_type ssp_rng_threefry = {
  rng_cxx_init<threefry_T>,
  rng_cxx_release<threefry_T>,
  rng_cxx_set<threefry_T>,
  rng_cxx_get<threefry_T>,
  rng_cxx_discard<threefry_T>,
  rng_cxx_read<threefry_T>,
  rng_cxx_write<threefry_T>,
  rng_cxx_entropy<threefry_T>,
  threefry_T::min(),
  threefry_T::max(),
  sizeof(threefry_T),
  16
};

#ifdef WITH_R123_AES
/* aes generator */
const struct ssp_rng_type ssp_rng_aes = {
  rng_cxx_init<aes_T>,
  rng_cxx_release<aes_T>,
  rng_cxx_set<aes_T>,
  rng_cxx_get<aes_T>,
  rng_cxx_discard<aes_T>,
  rng_cxx_read<aes_T>,
  rng_cxx_write<aes_T>,
  rng_cxx_entropy<aes_T>,
  aes_T::min(),
  aes_T::max(),
  sizeof(aes_T),
  16
};
#endif

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static char
check_type(const struct ssp_rng_type* type)
{
  ASSERT(type);
  return NULL != type->init
      && NULL != type->release
      && NULL != type->set
      && NULL != type->get
      && NULL != type->read
      && NULL != type->write
      && NULL != type->entropy
      && type->min <= type->max;
}

static void
rng_release(ref_T* ref)
{
  struct ssp_rng* rng;
  ASSERT(ref);
  rng = CONTAINER_OF(ref, struct ssp_rng, ref);
  if(rng->state) {
    if (rng->type.release) {
      rng->type.release(rng->state);
    }
    MEM_RM(rng->allocator, rng->state);
  }
  MEM_RM(rng->allocator, rng);
}

/*******************************************************************************
 * Exported function
 ******************************************************************************/
res_T
ssp_rng_create
  (struct mem_allocator* mem_allocator,
   const struct ssp_rng_type* type,
   struct ssp_rng** out_rng)
{
  struct mem_allocator* allocator;
  struct ssp_rng* rng = NULL;
  size_t log2r;
  res_T res = RES_OK;

  if(!type || !out_rng || !check_type(type)) {
    res = RES_BAD_ARG;
    goto error;
  }

  allocator = mem_allocator ? mem_allocator : &mem_default_allocator;
  rng = (struct ssp_rng*)MEM_CALLOC(allocator, 1, sizeof(struct ssp_rng));
  if(!rng) {
    res = RES_MEM_ERR;
    goto error;
  }
  rng->allocator = allocator;
  ref_init(&rng->ref);

  rng->state =
    MEM_ALLOC_ALIGNED(rng->allocator, type->sizeof_state, type->alignof_state);
  if(!rng->state) {
    res = RES_MEM_ERR;
    goto error;
  }
  res = type->init(allocator, rng->state);
  if(res != RES_OK) goto error;
  rng->type = *type;

  /* Precompute some values for the canonical distribution */
  rng->r = (long double)rng->type.max - (long double)rng->type.min + 1.0L;
  log2r = (size_t)(std::log(rng->r) / std::log(2.0L));
  rng->dbl_k = std::max<size_t>(1UL, (53 + log2r - 1UL) / log2r);
  rng->flt_k = std::max<size_t>(1UL, (24 + log2r - 1UL) / log2r);

exit:
  if(out_rng) *out_rng = rng;
  return res;
error:
  if(rng) {
    SSP(rng_ref_put(rng));
    rng = NULL;
  }
  goto exit;
}

res_T
ssp_rng_ref_get(struct ssp_rng* rng)
{
  if(!rng) return RES_BAD_ARG;
  ref_get(&rng->ref);
  return RES_OK;
}

res_T
ssp_rng_ref_put(struct ssp_rng* rng)
{
  if(!rng) return RES_BAD_ARG;
  ref_put(&rng->ref, rng_release);
  return RES_OK;
}

res_T
ssp_rng_get_type(const struct ssp_rng* rng, struct ssp_rng_type* type)
{
  if(!rng || !type) return RES_BAD_ARG;
  *type = rng->type;
  return RES_OK;
}

uint64_t
ssp_rng_get(struct ssp_rng* rng)
{
  if(!rng) FATAL("The Random Number Generator is NULL\n");
  return rng->type.get(rng->state);
}

uint64_t
ssp_rng_uniform_uint64
  (struct ssp_rng* rng, const uint64_t lower, const uint64_t upper)
{
  if(!rng) FATAL("The Random Number Generator is NULL\n");
  rng_cxx rng_cxx(*rng);
  return RAN_NAMESPACE::uniform_int_distribution<uint64_t>(lower, upper)(rng_cxx);
}

double
ssp_rng_uniform_double
  (struct ssp_rng* rng, const double lower, const double upper)
{
  if(!rng) FATAL("The Random Number Generator is NULL\n");
  rng_cxx rng_cxx(*rng);
  return RAN_NAMESPACE::uniform_real_distribution<double>(lower, upper)(rng_cxx);
}

float
ssp_rng_uniform_float
  (struct ssp_rng* rng, const float lower, const float upper)
{
  if(!rng) FATAL("The Random Number Generator is NULL\n");
  rng_cxx rng_cxx(*rng);
  return RAN_NAMESPACE::uniform_real_distribution<float>(lower, upper)(rng_cxx);
}

double
ssp_rng_canonical(struct ssp_rng* rng)
{
  if(!rng) FATAL("The Random Number Generator is NULL\n");
#if 0
  rng_cxx rng_cxx(*rng);
  return RAN_NAMESPACE::generate_canonical
    <double, std::numeric_limits<double>::digits>(rng_cxx);
#else
  /* Optimized version */
  size_t k = rng->dbl_k;
  double sum = 0;
  double tmp = 1;
  for(; k !=0; --k) {
    sum += (double)(rng->type.get(rng->state) - rng->type.min) * tmp;
    tmp = (double)(tmp*rng->r);
  }
  return sum/tmp;
#endif
}

float
ssp_rng_canonical_float(struct ssp_rng* rng)
{
  if(!rng) FATAL("The Random Number Generator is NULL\n");
  /* The C++ standard library does not ensure that the generated single
   * precision floating point number is effectively canonical, i.e. it may be
   * equal to 1. Use a hand made workaround to handle this bug. */
#if 0
  rng_cxx rng_cxx(*rng);
  return RAN_NAMESPACE::generate_canonical
    <float, std::numeric_limits<float>::digits>(rng_cxx);
#else
  float r;
  do {
    /* Optimised version of the generate_canonical function */
    size_t k = rng->flt_k;
    float sum = 0;
    float tmp = 1;
    for(; k !=0; --k) {
      sum += (float)(rng->type.get(rng->state) - rng->type.min) * tmp;
      tmp = (float)(tmp*rng->r);
    }
    r = sum/tmp;
  } while(r >= 1);
  return r;
#endif
}

uint64_t
ssp_rng_min(struct ssp_rng* rng)
{
  if(!rng) FATAL("The Random Number Generator is NULL\n");
  return rng->type.min;
}

uint64_t
ssp_rng_max(struct ssp_rng* rng)
{
  if(!rng) FATAL("The Random Number Generator is NULL\n");
  return rng->type.max;
}

res_T
ssp_rng_set(struct ssp_rng* rng, const uint64_t seed)
{
  if(!rng) return RES_BAD_ARG;
  return rng->type.set(rng->state, seed);
}

res_T
ssp_rng_read(struct ssp_rng* rng, FILE* stream)
{
  if(!rng || !stream) return RES_BAD_ARG;
  return rng->type.read(rng->state, stream);
}

res_T
ssp_rng_write(const struct ssp_rng* rng, FILE* stream)
{
  if(!rng || !stream) return RES_BAD_ARG;
  return rng->type.write(rng->state, stream);
}

double
ssp_rng_entropy(const struct ssp_rng* rng)
{
  if (!rng) FATAL("The Random Number Generator is NULL\n");
  return rng->type.entropy(rng->state);
}

res_T
ssp_rng_discard(struct ssp_rng* rng, uint64_t n)
{
  if (!rng) FATAL("The Random Number Generator is NULL\n");
  return rng->type.discard(rng->state, n);
}
